#!/usr/bin/env python

"""
Thin wrapper around the "az" command line interface (CLI) for use
with LocalStack.

The "azlocal" CLI allows you to easily interact with your local Azure services
without having to configure anything.

Example:
Instead of the following command ...
HTTPS_PROXY=... REQUESTS_CA_BUNDLE=... az storage account list
... you can simply use this:
azlocal storage account list

Options:
  Run "azlocal help" for more details on the Azure CLI subcommands.
"""

import logging
import os
import subprocess
import sys
from pathlib import Path

from .constants import AZURE_CONFIG_DIR_ENV
from .shared import prepare_environment, get_proxy_endpoint, get_proxy_env_vars, AZURE_CONFIG_DIR, run_in_background


DEFAULT_CLOUD_NAME = "AzureCloud"
CUSTOM_CLOUD_NAME = "LocalStack"

LOG = logging.getLogger(__name__)
logging.basicConfig(
    level=os.environ.get('LOGLEVEL', 'WARNING').upper(),
    format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
)


def usage():
    print(__doc__.strip())


def run_in_current_process(cmd, env = None):
    """
    Replaces this process with the AZ CLI process, with the given command and environment
    """
    os.execvpe(cmd[0], cmd, env)


def run(cmd, hide_errors: bool = False):
    LOG.debug(f"Executing {cmd}...")
    process = subprocess.run(cmd, capture_output = True)
    if process.stdout:
        LOG.debug(f"Output of {cmd}:")
        LOG.debug(process.stdout)
    if process.stderr and not hide_errors:
        if process.stderr.startswith(b"WARNING"):
            LOG.warning(process.stderr)
        else:
            LOG.error(f"Error while running {cmd}:")
            LOG.error(process.stderr)
    return process.returncode


def main():
    if len(sys.argv) > 1 and sys.argv[1] == '-h':
        usage()

    if len(sys.argv) > 1 and sys.argv[1] == "start_interception":
        use_custom_cloud()
    elif len(sys.argv) > 1 and sys.argv[1] == "stop_interception":
        use_regular_cloud()
    else:
        # Backwards compatibility
        # The user could invoke the CLI as if it were an 'az' replacement:
        #    azlocal group list
        # If that's the case, we redirect the existing 'az' command via a proxy to our Emulator
        run_as_separate_process()


def use_custom_cloud():
    # Create the CustomCloud first
    create_custom_cloud()
    # Configure the CLI to use our CustomCloud
    run(["az", "cloud", "set", "--name", CUSTOM_CLOUD_NAME])
    LOG.info(f"AZ CLI now uses CustomCloud={CUSTOM_CLOUD_NAME}")
    # Our endpoint is not whitelisted, so we need to disable the endpoint validation
    run(["az", "config", "set", "core.instance_discovery=false"])
    # Login
    login_failed = run(["az", "login", "--service-principal", "-u", "any-app", "-p", "any-pass", "--tenant", "anytenant"])
    # This might fail if the Emulator is not running
    # If that's the case, we should stop immediately
    if login_failed:
        exit(login_failed)
    LOG.info(f"AZ CLI is now logged in against CustomCloud={CUSTOM_CLOUD_NAME}")


def use_regular_cloud():
    # Configure the endpoint to use the default cloud
    run(["az", "cloud", "set", "--name", DEFAULT_CLOUD_NAME])
    # Re-enable instance discovery
    run(["az", "config", "set", "core.instance_discovery=true"])
    LOG.info(f"AZ CLI now uses Cloud={DEFAULT_CLOUD_NAME}")


def create_custom_cloud():
    custom_cloud_return_code = run(["az", "cloud", "show", "--name", CUSTOM_CLOUD_NAME], hide_errors=True)
    if custom_cloud_return_code == 0:
        # Exists
        return

    emulator_endpoint = os.environ.get("LOCALSTACK_HOST") or 'https://localhost.localstack.cloud:4566'
    # We currently only register the four necessary endpoints
    # We may want to expand this in the future and also register data plane endpoints
    # For example '--suffix-keyvault-dns' or '--suffix-storage-endpoint'
    run(["az", "cloud", "register", "--name", CUSTOM_CLOUD_NAME,
         "--endpoint-resource-manager", emulator_endpoint,
         "--endpoint-management", emulator_endpoint,
         "--endpoint-active-directory", emulator_endpoint,
         "--endpoint-active-directory-resource-id", emulator_endpoint,
         "--endpoint-active-directory-graph-resource-id", emulator_endpoint,
    ])
    LOG.info(f"Created CustomCloud={CUSTOM_CLOUD_NAME}")


def run_as_separate_process():
    """
    Constructs a command line string and calls "az" as an external process.
    """

    cmd_args = list(sys.argv)
    cmd_args[0] = 'az'
    if ("--help" in cmd_args) or ("--version" in cmd_args):
        # Early exit - if we only want to know the version/help, we don't need LS to be running
        run_in_current_process(cmd_args, None)
        return

    proxy_endpoint = get_proxy_endpoint()

    env_dict = prepare_environment(proxy_endpoint)

    env_dict[AZURE_CONFIG_DIR_ENV] = AZURE_CONFIG_DIR
    if not os.path.exists(AZURE_CONFIG_DIR):
        # Create the config directory
        Path(AZURE_CONFIG_DIR).mkdir(parents=True, exist_ok=True)

        # Prepare necessary arguments to ensure `az ..` commands are run against this config directory
        az_args_list = [f"{key}={val}" for key, val in get_proxy_env_vars(proxy_endpoint).items()]
        az_args_list.append(f"{AZURE_CONFIG_DIR_ENV}={AZURE_CONFIG_DIR}")
        az_arg = " ".join(az_args_list)

        # Turn off telemetry
        survey_command = f"{az_arg} az config set output.show_survey_link=no"
        run_in_background(survey_command)
        telemetry_command = f"{az_arg} az config set core.collect_telemetry=false"
        run_in_background(telemetry_command)

        # Login to ensure the config directory has credentials
        login_command = f"{az_arg} az login --service-principal -u any-app -p any-pass --tenant any-tenant"
        run_in_background(login_command)

    # Hijack the login command
    # When creating our custom config dir, we automatically log in - so this is not necessary anymore
    if len(cmd_args) == 2 and cmd_args[1] == "login":
        print("Login Succeeded")
        return

    # Hijack the ACR login command
    if len(cmd_args) > 1 and cmd_args[1] == "acr" and "login" in cmd_args:
        print("Login Succeeded")
        return

    # run the command
    run_in_current_process(cmd_args, env_dict)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
Thin wrapper around the "azd" command line interface (CLI) for use
with LocalStack.

The "azdlocal" CLI allows you to easily spin up your IaC against LocalStack
without having to configure anything.

Example:
Instead of the following command ...
HTTPS_PROXY=... SSL_CERT_FILE=... azd up
... you can simply use this:
azdlocal up

Options:
  Run "azdlocal help" for more details on the Azure CLI subcommands.
"""

import os
import sys

from .shared import check_proxy_is_running, get_proxy_endpoint, prepare_environment


def usage():
    print(__doc__.strip())


def run(cmd, env):
    """
    Replaces this process with the AZ CLI process, with the given command and environment
    """
    os.execvpe(cmd[0], cmd, env)


def main():
    if len(sys.argv) > 1 and sys.argv[1] == '-h':
        return usage()
    run_as_separate_process()


def run_as_separate_process():
    """
    Constructs a command line string and calls "az" as an external process.
    """

    cmd_args = list(sys.argv)
    cmd_args[0] = 'azd'
    if ("help" in cmd_args) or ("version" in cmd_args):
        # Early exit - if we only want to know the version/help, we don't need LS to be running
        run(cmd_args, None)
        return

    proxy_endpoint = get_proxy_endpoint()
    check_proxy_is_running(proxy_endpoint)

    env_dict = prepare_environment(proxy_endpoint)

    # Hijack the login command to automatically login
    if len(cmd_args) > 1 and cmd_args[1] == "auth" and "--help" not in cmd_args:
        cmd_args = ["azd", "auth", "login", "--client-id", "any-app", "--client-secret", "any-pass", "--tenant-id", "any-tenant"]

    # run the command
    run(cmd_args, env_dict)


if __name__ == '__main__':
    main()



#!/usr/bin/env python

"""
Thin wrapper around the "terraform" command line interface (CLI) for use
with LocalStack.

The "azlocal" CLI allows you to easily interact with your local Azure services
without having to configure anything.

Example:
Instead of the following command ...
HTTPS_PROXY=... REQUESTS_CA_BUNDLE=... az storage account list
... you can simply use this:
azlocal storage account list

Options:
  Run "azlocal help" for more details on the Azure CLI subcommands.
"""

import os
import sys
from pathlib import Path

from .constants import AZURE_CONFIG_DIR_ENV
from .shared import AZURE_CONFIG_DIR, check_proxy_is_running, get_proxy_endpoint, prepare_environment, run_in_background


def usage():
    print(__doc__.strip())


def run(cmd, env):
    """
    Replaces this process with the Terraform CLI process, with the given command and environment
    """
    os.execvpe(cmd[0], cmd, env)


def main():
    if len(sys.argv) > 1 and sys.argv[1] == '-h':
        return usage()
    run_as_separate_process()


def run_as_separate_process():
    """
    Constructs a command line string and calls "terraform" as an external process.
    """

    cmd_args = list(sys.argv)
    cmd_args[0] = 'terraform'
    if ("--help" in cmd_args) or ("version" in cmd_args):
        # Early exit - if we only want to know the version/help, we don't need LS to be running
        run(cmd_args, None)
        return

    proxy_endpoint = get_proxy_endpoint()
    check_proxy_is_running(proxy_endpoint)

    if "init" not in cmd_args:
        # Init downloads all dependencies - no point in going through our proxy
        env_dict = prepare_environment(proxy_endpoint)
    else:
        env_dict = os.environ

    # Configure a custom configuration directory
    env_dict[AZURE_CONFIG_DIR_ENV] = AZURE_CONFIG_DIR
    if not os.path.exists(AZURE_CONFIG_DIR):
        # Create the config directory
        Path(AZURE_CONFIG_DIR).mkdir(parents=True, exist_ok=True)

        args_list = [f"{key}={val}" for key, val in env_dict.items()]
        args_list.append(f"{AZURE_CONFIG_DIR_ENV}={AZURE_CONFIG_DIR}")
        args = " ".join(args_list)

        # Initial login to ensure auth details are part of the configuration
        login_command = f"{args} az login --service-principal -u any-app -p any-pass --tenant any-tenant"
        run_in_background(login_command)

    # run the command
    run(cmd_args, env_dict)


if __name__ == '__main__':
    main()



import logging
import os
import sys
import inspect
import warnings
from pathlib import Path
from time import time

ENV_AZURE_AUTHORITY_HOST = "AZURE_AUTHORITY_HOST"

LOG = logging.getLogger(__name__)


def get_custom_host() -> str:
    return os.environ.get("LS_HOST") or "https://localhost.localstack.cloud:4566"


class PythonLocalSdk:

    def __init__(self):
        self._started = False

        self._clients_to_mock = self._find_azure_mgmt_clients()
        self._default_worldwide_endpoint: str | None = None
        self._existing_authority_host: str | None = None

    def start_interception(self):
        """
        Start intercepting all Azure requests.

         Invoking this method will:
          - inspect the current modules
          - load all `azure.mgmt` modules
          - find the latest client in every module
          - update the client to point to the LocalStack host

        Use the environment variable `LS_HOST` to configure the host.

        .. warning
              This only affects Clients that are created after this method is invoked

        :return:
        """
        if self._started:
            return

        # Set custom login endpoint
        # There are two ways that Azure SDK Clients determine which endpoint to call to login:
        #  - Using an environment variable
        #  - Using hardcoded values in the 'msal' module
        # Which method is used depends on the SDK, so we override both to be sure
        self._existing_authority_host = os.environ.get(ENV_AZURE_AUTHORITY_HOST)
        os.environ[ENV_AZURE_AUTHORITY_HOST] = get_custom_host()
        try:
            from msal import authority

            self._default_worldwide_endpoint = authority.WORLD_WIDE
            authority.WORLD_WIDE = get_custom_host().replace("https://", "")
        except ImportError:
            warnings.warn("Unable to import 'msal' - Authorization calls will not be intercepted")

        # Use Custom Client
        for client in self._clients_to_mock:
            self.prepare_client_for_interception(client)

    def stop_interception(self):
        if not self._started:
            return

        # Instrument client to stop interception
        for client in self._clients_to_mock:
            self.revert_client_interception(client)

        # Revert the login endpoints
        if self._existing_authority_host:
            os.environ[ENV_AZURE_AUTHORITY_HOST] = self._existing_authority_host
        if self._default_worldwide_endpoint:
            from msal import authority

            authority.WORLD_WIDE = self._default_worldwide_endpoint

        self._started = False

    def _find_azure_mgmt_clients(self) -> list:
        """
        Finds a list of all SDK clients that are defined in the `azure.mgmt` namespace.

        :return: A list of classes: [CosmosManagementClient, StorageManagementClient, ..]
        """

        from azure import mgmt  # noqa
        azure_mod = sys.modules["azure.mgmt"]
        azure_path = azure_mod.__path__

        clients = []

        total_start_time = time()

        for p in azure_path:
            mgmt_modules = sorted(os.listdir(p))

            for module_name in mgmt_modules:
                module_start_time = time()
                module = __import__(f"azure.mgmt.{module_name}", fromlist=["__init__"])
                LOG.debug(f"PythonSDK: importing %s took: \t%s s", module_name, (time() - module_start_time))

                if hasattr(module, '__all__'):
                    for attr_name in module.__all__:
                        if attr_name.endswith("Client"):
                            clients.append(getattr(module, attr_name))
                else:
                    # some modules (datalake, rdbms, maybe others) use have several clients in subdirectories
                    sub_modules = sorted(os.listdir(os.path.join(p, module_name)))

                    for submodule_name in sub_modules:
                        if not Path(os.path.join(p, module_name, submodule_name)).is_dir():
                            continue
                        submodule = __import__(f"azure.mgmt.{module_name}.{submodule_name}", fromlist=["__init__"])

                        if hasattr(submodule, '__all__'):
                            for attr_name in submodule.__all__:
                                if attr_name.endswith("Client"):
                                    clients.append(getattr(submodule, attr_name))

        LOG.debug(f"PythonSDK: finding all clients took \t%s s", (time() - total_start_time))

        return clients

    @staticmethod
    def prepare_client_for_interception(client) -> None:
        # There are three ways for a client to determine which endpoint to call
        # 1. Provided by the user: AzureServiceClient(base_url="..")
        # 2. Determined by the AzureCloud that is configured (env variable AZURE_CLOUD=AZURE_PUBLIC_CLOUD/AZURE_CHINA_CLOUD
        # 3. Use a hardcoded default
        #
        # Solutions:
        # 1. Asking the user to configure the base_url manually. Very laborious and prone to errors
        # 2. Override the AzureCloud. This is not feasible for a few reasons:
        #    - We can pass in an env variable, but there are only three Azure Clouds (Public/US GOV/China)
        #    - We cannot override the available AzureClouds, as that is an Enum (which is immutable by design)
        #    - We cannot override the method used to determine the endpoint
        #      1. Location of this method: from azure.mgmt.core.tools import get_arm_endpoints
        #      2. This is typically already imported at the top, so each client already has a direct reference to the `get_arm_endpoints` method.
        #         Overriding the method in the `azure.mgmt.core.tools`-module has therefore no effect.
        #
        # 3. We _can_ override the hardcoded default - which is what we do here

        # Determine (and store) the existing default values for the `__init__`-method
        # This is either None or a tuple
        # If a parameter does not have a default value, it is not part of the tuple - so the length of this list is not guaranteed the same as the number of parameters
        # __init__(a, b = None, c = "sth") --> _defaults__ == (None, "sth")
        client.__init__.__olddefaults = client.__init__.__defaults__

        if not client.__init__.__olddefaults:
            return

        # Get the default values as a dictionary
        # We skip the 'empty' Parameters, so parameters without a default value
        # __init__(a, b = None, c = "sth") --> [("b", None), ("c", "sth")]
        client_signature = inspect.signature(client.__init__)
        init_params = [(name, value.default) for name, value in client_signature.parameters.items() if value.default != inspect.Parameter.empty]

        # Construct our own tuple of default values, only overriding the `base_url`
        updated_default_values = tuple(get_custom_host() if name == "base_url" else value for name, value in init_params)

        client.__init__.__defaults__ = updated_default_values

    @staticmethod
    def revert_client_interception(client) -> None:
        client.__init__.__defaults__ = client.__init__.__olddefaults



import json
import os
import tempfile
import subprocess
import urllib.error
import urllib.request
from typing import Optional
from pathlib import Path

import certifi


HOME_DIR = Path.home()
AZURE_DIR = os.path.join(HOME_DIR, ".localstack/azure")
AZURE_CONFIG_DIR = os.path.join(AZURE_DIR, "az_config")

if not os.path.exists(AZURE_DIR):
    Path(AZURE_DIR).mkdir(parents=True, exist_ok=True)


def get_proxy_endpoint() -> str:
    ls_host = os.environ.get("LOCALSTACK_HOST") or 'http://localhost:4566'
    with urllib.request.urlopen(f"{ls_host}/_localstack/proxy") as req:
        proxy_details = json.loads(req.read().decode("utf-8"))
        proxy_port = proxy_details["proxy_port"]
    return f'http://localhost:{proxy_port}'


def get_default_proxy_certificate_location(proxy_endpoint):
    certificate_path = os.path.join(AZURE_DIR, "ca.crt")
    if not os.path.exists(certificate_path):
        certificate_bytes = _get_certificate_bytes(proxy_endpoint)
        Path(certificate_path).write_bytes(certificate_bytes)
    return certificate_path


def get_custom_proxy_certificate_location(proxy_endpoint) -> str:
    with tempfile.NamedTemporaryFile(delete=False) as tmp:
        tmp.write(_get_certificate_bytes(proxy_endpoint))
    return tmp.name


def _get_certificate_bytes(proxy_endpoint: str) -> bytes:
    # Stock CA certificates
    certificate_bytes = Path(certifi.where()).read_bytes()

    # Custom CA certificate
    with urllib.request.urlopen(f"{proxy_endpoint}/_localstack/certs/ca/LocalStack_LOCAL_Root_CA.crt") as cert:
        certificate_bytes += cert.read()
    return certificate_bytes


def check_proxy_is_running(proxy_endpoint: str):
    try:
        assert urllib.request.urlopen(f"{proxy_endpoint}/_localstack/health").status == 200
    except (AssertionError, urllib.error.URLError) as e:
        raise Exception("Make sure LocalStack is running")


def prepare_environment(proxy_endpoint: str):
    # prepare env vars
    env_dict = os.environ.copy()
    env_dict.update(get_proxy_env_vars(proxy_endpoint))

    # update environment variables in the current process
    os.environ.update(env_dict)
    return env_dict



def get_proxy_env_vars(proxy_endpoint: str, certificate_path: Optional[str] = None):
    certificate_path = certificate_path or get_default_proxy_certificate_location(proxy_endpoint)

    return {
        "HTTP_PROXY": proxy_endpoint,
        "HTTPS_PROXY": proxy_endpoint,
        "REQUESTS_CA_BUNDLE": certificate_path,
        "SSL_CERT_FILE": certificate_path,
    }


def run_in_background(command):
    p = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    p.wait()
    log_az_output(p)


def log_az_output(p):
    for line in p.stdout.readlines():
        # TODO: Log this cleanly
        #  Just not to stdout, as we should only print the output of the actual command that the user runs
        pass
